//+------------------------------------------------------------------+
//|                                            moex-forts-splice.mq5 |
//|                                    Copyright (c) 2025, Marketeer |
//|                           https://www.mql5.com/ru/articles/16964 |
//+------------------------------------------------------------------+
#property description "Splicing of futures from MOEX's FORTS market"
#property copyright "Copyright (c) 2025, Marketeer"
#property link "https://www.mql5.com/ru/articles/16964"
#property script_show_inputs

#include "moexcore.mqh"
#include "moex2mql5.mqh"
#include "moexref.mqh"

input group "MOEX Asset Selector"
const MOEX_ENGINES Engine = MOEX_ENGINES::futures; // implied by/deduced from market
const MOEX_MARKETS Market = futures·forts;
const MOEX_BOARDS Board = forts·RFUD; // only one board in FORTS
input string AssetCodes = "GOLD,SILV,PLT,PLD";
input string DefaultCurrency = "USD";

input group "History"
input datetime From = 0; // From (1970 = very beginning)
input MOEX_INTERVAL Timeframe = M1; // Timeframe (M1 recommended)

input group "Auxiliary"
input bool Logging = false;
input bool Dumping = true;
input bool Test = false;
input uint Timeout = 5; // Timeout (seconds)
input uint Delay = 100; // Delay (between paging requests, ms)
input string Parameters = "iss.meta=off"; // Global Parameters (such as iss.meta=on|off, etc)
// save/load MqlRates for faster resuming after occasional interruption by an error
input bool CacheRates = false;

enum ENUM_SPLICING
{
   NEXT_EXPIRING,   // Nearest expiring
   VOLUME_WEIGHTED, // Volume weighted
   TIME_WEIGHTED,   // Remaining days weighted
};

input group "Splicing"
input ENUM_SPLICING SplicingAlgorithm = NEXT_EXPIRING;
input uint AveragingPeriod = 0; // AveragingPeriod (min)


const string algo[] = {"N", "V", "R"};

// Examples of starting requests for getting lists of futures:
// https://iss.moex.com/iss/engines/futures/markets/forts/boards/RFUD/securities.json?security_collection=futures_forts_commodity&sort_column=lasttradedate
// https://iss.moex.com/iss/engines/futures/markets/forts/boards/RFUD/securities.json?security_collection=futures_forts_commodity&sort_column=openposition&sort_order=desc

bool CreateCustomSymbol(const string ticker, MOEX_BOARDS b, MOEX_MARKETS m, MOEX_ENGINES e, const string c, const string name = NULL, const string desc = NULL)
{
   Print("Creating new custom symbol ", ticker, name != NULL ? " (" + name + ")" : "");
   string market, engine;
   Moex::ResolveMoexNames(b, m, e, market, engine);
   
   // Single ticker extensive description from iss/securities/XYZ
   JsValue *about = iss.securities.security(ticker)._get("lang=en"); // use lang=en to detect 'unit' currency
   // if(Dumping) DumpJsonToFile(StringFormat("MOEX/about-%s.json", UseTicker), about);
  
   // Specs for ticker /iss/engines/[engine]/markets/[market]/boards/[board]/securities/XYZ
   JsValue *spec = iss.engines.security(ticker, b, m, e)._get();
   
   if(spec.t == JS_OBJECT) // not an error or null
   {
      if(about.t == JS_OBJECT) // augment specs with description
      {
         spec.merge(about, true);
      }
      
      if(Dumping) DumpJsonToFile(StringFormat("MOEX/symbol-%s.json", ticker), spec);
      // Create custom symbol at engine/market/ticker path
      // TODO: disable expiration property
      if(Moex::CreateCustomSymbol(spec, name == NULL ? ticker : name, market, engine, c, desc))
      {
         Print("Successfully created");
      }
      else
      {
         Print("Failed with error: ", _LastError);
      }
      return true;
   }
   return false;
}

const long MONTH_LONG = 60 * 60 * 24 * 30;
const long DAY_LONG = 60 * 60 * 24;

struct TickerHistory
{
   string ticker;
   datetime expiration;
   MqlRates rates[];
};

void OnStart()
{
   string codes[];
   if(!StringSplit(AssetCodes, ',', codes)) return;
   
   iss.link.applySettings(MoexLink::Settings(Logging, Dumping, Test, Timeout, Delay, Parameters));
   
   AutoPtr<JsValue> workset;
   for(int i = 0; i < ArraySize(codes) && !IsStopped(); i++)
   {
      // https://iss.moex.com/iss/statistics/engines/futures/markets/forts/series.json
      // normally returned sorted from current to historic
      // this is where we are strictly bound to FORTS
      JsValue *assets = iss.statistics.futuresFortsSeries()._get(StringFormat("asset_code=%s&show_expired=1", codes[i]));
      static AutoPtr<JsValue> clmn = JsValue::toUniformCase(assets["series"]["columns"]);
      // find out indices of columns
      static int secid = clmn[].indexOf("secid");
      static int name = clmn[].indexOf("name");
      static int start_date = clmn[].indexOf("start_date");
      static int expiration_date = clmn[].indexOf("expiration_date");
      
      JsValue *array = assets["series"]["data"];
      workset = new JsValue(JS_ARRAY);
      datetime prev2 = 0;
      for(int j = 0; j < array.length(); j++)
      {
         // accept normal futures, excluding calendar spreads and early incomplete contracts
         if(array[j][secid].s != array[j][name].s &&
            array[j][expiration_date].get<datetime>() - array[j][start_date].get<datetime>() >= MONTH_LONG * 5)
         {
            workset[].put(array[j]);
            array.drop(j);
            
            if(prev2 && prev2 < array[j][expiration_date].get<datetime>())
            {
               Alert(StringFormat("Retrospective order is broken @ %s", array[j][secid].s));
               DebugBreak();
            }
            
            prev2 = array[j][expiration_date].get<datetime>();
         }
      }
      
      if(Dumping) DumpJsonToFile(StringFormat("MOEX/futures/series_%s.json", codes[i]), workset[]);
      
      const string counter = StringFormat("%d/%d assets,", i + 1, ArraySize(codes));
      int requests = 0;
      
      TickerHistory history[];
      long length = 0;
      string compound = codes[i] + ".iss" + algo[SplicingAlgorithm] + (AveragingPeriod > 1 ? (string)AveragingPeriod : "");

      bool isCustom;
      if(SymbolExist(compound, isCustom)) continue;
      
      const int m = workset[].length();
      for(int j = 0; j < m && !IsStopped(); j++)
      {
         const string ticker = workset[][j][secid].s;
         
         if(!j)
         {
            // request ISS for specs and create custom symbol for them
            if(!CreateCustomSymbol(ticker, Board, Market, Engine, DefaultCurrency, compound, "Spliced " +
               EnumToString(SplicingAlgorithm) + (AveragingPeriod > 1 ? " ave(" + (string)AveragingPeriod + ")" : ""))) continue;
         }
         
         const string subcounter = StringFormat(" %s %d/%d tickers, %d requests", ticker, j + 1, m, ++requests);
         Comment(counter + subcounter);

         TickerHistory th;
         th.ticker = ticker;
         th.expiration = StringToTime(workset[][j][expiration_date].s);
         
         bool downloaded = false;
         if(CacheRates && Test)
         {
            downloaded = LoadRates(ticker, th.rates) >= 0;
         }
         
         if(!downloaded)
         {
            // download all candles
            AutoPtr<JsValue> data = iss.engines.candles(ticker, Board, Market, Engine)._interval(Timeframe)._from(From)._start(MOEX_ALL)._get();
            if(data[].t >= JS_NULL) break;
            PASS(Moex::GetRates(data[], th.rates));
            downloaded = true;
         }
         else
         {
            downloaded = false;
         }
         
         if(CacheRates && Dumping && downloaded)
         {
            SaveRates(ticker, th.rates);
         }
         
         PUSH(history, th);
         length += ArraySize(th.rates);
      }

      double dRates[][7];
      
      const int q = ArraySize(history);
      PrintFormat("Allocating %lld rates", length);
      ResetLastError();
      PRTF(ArrayResize(dRates, (int)length));
      if(_LastError)
      {
         Alert("Resize failed");
         return;
      }
      
      int cursor = 0;
      for(int j = 0; j < q; j++) // from the latest to the oldest
      {
         PrintFormat("[%d] %s %s", cursor, history[j].ticker, TimeToString(history[j].expiration, TIME_DATE));
         
         for(int k = ArraySize(history[j].rates) - 1; k >= 0; k--)
         {
            if(SplicingAlgorithm == NEXT_EXPIRING)
            {
               int jj = j;
               while(j < q - 1 && history[j + 1].expiration >= history[jj].rates[k].time / DAY_LONG * DAY_LONG)
               {
                  j++;
               }
               if(j != jj)
               {
                  PrintFormat("switch from %s at %s to %s", history[jj].ticker, (string)history[jj].rates[k].time, history[j].ticker);
                  j--;
                  break; // jump to next element/previous series at j+1
               }
            }
            
            dRates[cursor][0] = (double)history[j].rates[k].time;
            dRates[cursor][1] = history[j].rates[k].open;
            dRates[cursor][2] = history[j].rates[k].high;
            dRates[cursor][3] = history[j].rates[k].low;
            dRates[cursor][4] = history[j].rates[k].close;
            dRates[cursor][5] = (double)history[j].rates[k].tick_volume;
            dRates[cursor][6] = (double)history[j].expiration;
            cursor++;
         }
      }
      
      ArrayResize(dRates, cursor);
      PrintFormat("Sorting %lld rates...", ArraySize(dRates));
      ArraySort(dRates);
      Print("Sorted");
      
      MqlRates combined[];
      // const long alloc = length * 6 / q;
      // Print("Size estimate: ", alloc);
      datetime prev = (datetime)((long)dRates[0][0] / 60 / fmax(AveragingPeriod, 1) * fmax(AveragingPeriod, 1) * 60);
      int inBarCount = 0; // number of input bars during averaging
      const int n = ArrayRange(dRates, 0);
      for(int j = 0; j < n; j++)
      {
         const datetime dt = (datetime)((long)dRates[j][0] / 60 / fmax(AveragingPeriod, 1) * fmax(AveragingPeriod, 1) * 60);
         
         if((dt != prev) || j == n - 1)
         {
            long total = 0;
            double scale = 0;
            for(int k = 1; k <= inBarCount; k++)
            {
               total += (long)dRates[j - k][5];
               if(SplicingAlgorithm == TIME_WEIGHTED) scale += 1.0 / sqrt((dRates[j - k][6] - dRates[j - k][0]) / DAY_LONG + 1);
            }
            
            if(SplicingAlgorithm != TIME_WEIGHTED) scale = (double)total;
            
            MqlRates rate = {};
            rate.time = prev;
            prev = dt;
            rate.tick_volume = total;
            
            for(int k = 1; k <= inBarCount; k++)
            {
               double subscale = (SplicingAlgorithm == TIME_WEIGHTED ? 1.0 / sqrt((dRates[j - k][6] - dRates[j - k][0]) / DAY_LONG + 1) : dRates[j - k][5]);
               
               rate.open += dRates[j - k][1] * subscale;
               rate.high += dRates[j - k][2] * subscale;
               rate.low += dRates[j - k][3] * subscale;
               rate.close += dRates[j - k][4] * subscale;
            }
            
            rate.open /= scale;
            rate.high /= scale;
            rate.low /= scale;
            rate.close /= scale;
            
            // TODO: NormalizeDouble();
            
            PUSH(combined, rate);
            
            inBarCount = 1;
         }
         else
         {
            if(SplicingAlgorithm == NEXT_EXPIRING && j && AveragingPeriod <= 1)
               DebugBreak();
            inBarCount++;
         }
      }
      Print("Final size: ", ArraySize(combined));
      // ArrayPrint(combined);
      
      // submit all candles
      ResetLastError();
      PRTF(CustomRatesReplace(compound, NULL, LONG_MAX, combined));
   }
   
   Comment("");
   
}

bool SaveRates(const string ticker, const MqlRates &rates[])
{
   return FileSave("MOEX/futures/mqlrates/" + ticker + ".mqlrates", rates);
}

long LoadRates(const string ticker, MqlRates &rates[])
{
   return FileLoad("MOEX/futures/mqlrates/" + ticker + ".mqlrates", rates);
}
//+------------------------------------------------------------------+
